/*
Copyright (c) 2013  V.Ayllón.  All rights reserved.
 
 Inspired by HP2xx:
 
 Parts Copyright (c) 1991 - 1994 Heinz W. Werntges.  All rights reserved.
 Parts Copyright (c) 1998-2002,2003 Martin Kroeker  All rights reserved.
 
 DiyouPCBRun Console is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
 to anyone for the consequences of using it or for whether it serves any
 particular purpose or works at all, unless he says so in writing.  Refer
 to the GNU General Public License, Version 2 or later, for full details.
 
 Everyone is granted permission to copy, modify and redistribute
 DiyouPCBRun, but only under the conditions described in the GNU General Public
 License.  A copy of this license is supposed to have been
 given to you along with DiyouPCBRun so you can know your rights and
 responsibilities.  It should be in a file named COPYING.  Among other
 things, the copyright notice and this notice must be preserved on all
 copies.
 
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
import java.io.*;
import java.io.IOException;

public class HpglParser {

  private String HPTEMPFILENAME = tempDirectory+ "/hptempfile.gcode";
  private float Z_UP_POSITION = 2;  // Z axis up position in mm
  private float DEFAULT_FEEDRATE = 600; // Default feedrate in mm/min

  private int X_AXIS = 0;
  private int Y_AXIS = 1;
  private int NUM_AXIS = 2;

  private float[] offset =  new float[NUM_AXIS];
  private float[] current_position = new float[NUM_AXIS];
  private float[] destination = new float[NUM_AXIS];
  private float current_Z_position = Z_UP_POSITION;
  private float feedrate = DEFAULT_FEEDRATE;

  private int MAX_PARSE_ERRORS = 5;

  private float P1X_default = 0.0;
  private float P1Y_default = 0.0;
  private  float P2X_default = 100.0;	
  private float P2Y_default = 100.0;

  private float PLOT_UNIT = 0.025f;

  private int  MAX_CMD_SIZE = 96;
  private int  MAX_PARAMETERS = 4;

  private char mybuffer[] = new char[255];

  // PA command parameters
  private int  X_COORD = 0;
  private int  Y_COORD = 1;

  // AA command parameters

  private int  I_COORD  = 0;
  private int  J_COORD  = 1;
  private int  ARC_THETA = 2;
  private int  SEG_ALPHA = 3;

  // IP command parameters

  private int  IP_P1X = 0;
  private int  IP_P1Y = 1;
  private int  IP_P2X = 2;
  private int  IP_P2Y = 3;

  // SC command parameters

  private int  SC_P1X=0;
  private int  SC_P1Y=1;
  private int  SC_P2X=2;
  private int  SC_P2Y=3;


  /* Supported HPGL commands, ASCII-coded as High-byte/low-byte int's */

  private int  AA	= 0x4141;
  private int  IN	= 0x494E;
  private int  IP	= 0x4950;
  private int  PA	= 0x5041;
  private int  PD	= 0x5044;
  private int  PU       = 0x5055;
  private int  SC	= 0x5343;
  private int  SP	= 0x5350;
  private int  VS	= 0x5653;
  private int  PW	= 0x5057;
  // Command error
  private int  PARSER_ERROR = 0xffff;


  private  float[] numpar = new float[MAX_PARAMETERS];
  private  char[] buffptr;
  private  int buffinx = 0;

  private  float[] P1 = new float[NUM_AXIS];	// Scaling points
  private  float[] P2 = new float[NUM_AXIS];
  private  float[] S1 = new float[NUM_AXIS];
  private  float[] S2 = new float[NUM_AXIS];
  private  float[] Q = new float[NUM_AXIS];

  //////////  Constructors & Methods

  HpglParser() {
    init();
  }

  /** init() - Init variables
   ** HP-GL commands, reads parameters (if expected), and initiates processing of these
   ** commands.
   **/
  public void init() {

    P1[X_COORD] =  P1X_default;
    P1[Y_COORD] =  P1Y_default;
    P2[X_COORD] =  P2X_default; 
    P2[Y_COORD] =  P2Y_default;

    S1[X_COORD] =  P1X_default;
    S1[Y_COORD] =  P1Y_default;
    S2[X_COORD] =  P2X_default; 
    S2[Y_COORD] =  P2Y_default;


    Q[X_COORD] = 1.; 		/* Delta-P/Delta-S: Initialized with first SC   */
    Q[Y_COORD] = 1.;
  }


  /**
   ** Hpgl2Gcode() - Convert a  hpgl file to a g-code file
   ** fileWidth and fileHeight in mm
   **/
  public String hpgl2Gcode(String inputFileName) throws IOException {
    String line;
    int command = 0;
    int unknowCmd = 0;
    int numLine = 0;
    // Init variables
    feedrate = DEFAULT_FEEDRATE;
    //
    // create and open files
    //
    BufferedReader inputFile = null;
    BufferedWriter outputFile = null;

    try {
      inputFile = new BufferedReader(new FileReader(inputFileName));
      outputFile = new BufferedWriter(new FileWriter(HPTEMPFILENAME));
      //
      // Read lines, parse them, convert to g-code and write to disk
      do {
        line = inputFile.readLine();

        if (line != null) {
          string2Char(line);
          command = parse(mybuffer);
          numLine++;

          switch(command) {
          case 0x5055: //PU:
            outputFile.write("G00 Z"+ Z_UP_POSITION +" F"+feedrate +"\n");
            break;
          case 0x5044: //PD:
            outputFile.write("G00 Z"+ 0 +" F"+feedrate +"\n");
            break;
          case 0x5041: //PA:
            outputFile.write("G01 X"+ numpar[X_AXIS] +" Y" + numpar[Y_AXIS] + " F"+feedrate +"\n");
            current_position[X_AXIS] = numpar[X_AXIS];
            current_position[Y_AXIS] = numpar[Y_AXIS];
            //System.out.println("X "+current_position[X_AXIS]+" Y:"+current_position[Y_AXIS]);
            break;
          case 0x4141: //AA:
            float theta = radians(numpar[ARC_THETA]); 
            offset[I_COORD] = numpar[I_COORD]-current_position[X_AXIS];    // I (x of center)
            offset[J_COORD] = numpar[J_COORD]-current_position[Y_AXIS];    // J (y of center)
            // System.out.println("I "+offset[I_COORD]+" J:"+offset[J_COORD]);
            if (numpar[ARC_THETA] < 0)  //Check if is clockwise or not
              outputFile.write("G03 I"+ offset[I_COORD] +" J" + offset[J_COORD] + " F"+feedrate +"\n");
            else
              outputFile.write("G02 I"+ offset[I_COORD] +" J" + offset[J_COORD] + " F"+feedrate +"\n");
            current_position[X_AXIS] = numpar[I_COORD] + (current_position[X_AXIS] - numpar[I_COORD]) * cos(theta) - (current_position[Y_AXIS] - numpar[J_COORD]) * sin(theta);
            current_position[Y_AXIS] = numpar[J_COORD] + (current_position[X_AXIS] - numpar[I_COORD]) * sin(theta) + (current_position[Y_AXIS] - numpar[J_COORD]) * cos(theta);
            // System.out.println("X "+current_position[X_AXIS]+" Y:"+current_position[Y_AXIS]);
            break;
          case 0x494E: //IN:    // Init and home
            // home
            outputFile.write("; File converted from HPGL \n");
            outputFile.write("G21 ; set units to millimeters \n");
            outputFile.write("G90 ; use absolute coordinates \n");
            outputFile.write("G28 ; Home axis \n");
            outputFile.write("G00 X0 Y0 Z"+ Z_UP_POSITION +" F"+feedrate +"\n");
            current_position[X_AXIS] = current_position[Y_AXIS] = 0.f;
            break;
          case 0x5653: //VS:
            feedrate = numpar[0]*60; // Convert to mm/min
            break;
          case 0x5057: //PW:
            break;
          case 0x5350: //SP:    // Select pen 
          case 0x4950: //IP:    // Init point
          case 0x5343: //SC:    // Scale all of them processed inside parser
            break;
          default:
            unknowCmd++;
            if (unknowCmd > MAX_PARSE_ERRORS)
              throw new IOException("Max number of HPGL parser errors reached");
            break;
          } // Switch
        } // line != null
      } 
      while (line != null);
    } 
    catch (IOException e) {
      if (outputFile != null)
        outputFile.close();
      if (inputFile != null)
        inputFile.close();
      throw new IOException("Error converting HPGL file: " + e.toString());
    }
    // Close files
    //System.out.println("Close HPGL Parser...");
    if (outputFile != null)
      outputFile.close();
    if (inputFile != null)
      inputFile.close();

    // Check if empty file
    if (numLine == 0)
      throw new IOException("Error converting HPGL file. The file is empty");

    // Return the new file name
    return HPTEMPFILENAME;
  }

  /** parse() - Reads the input stream character-by-character, identifies
   ** HP-GL commands, reads parameters (if expected), and initiates processing of these
   ** commands.
   **/
  private int parse(char[] buffer) {
    buffptr = buffer;
    buffinx = 0;
    int c;
    int cmd;

    while ( (c = buffer[buffinx++]) != '\0') {
      if ((c < 'A') || (c > 'z')
        || ((c > 'Z') && (c < 'a')))
        continue;
      cmd = c << 8;
      if ((c = buffptr[buffinx++]) == '\0')
        break;	// Partial command
      if ((c < 'A') || (c > 'z')
        || ((c > 'Z') && (c < 'a'))) {
        continue;
      }
      cmd |= (c & 0xFF);
      cmd &= 0xDFDF; // Pass to upper
      //			print(pbuffer);
      return parseCommand(cmd);
    }
    return PARSER_ERROR;
  }

  /**
   ** parseCommand() - each command consists of 2 characters
   ** This function returns the command detected.
   **/
  private int parseCommand(int command) {

    //			print(pbuffer);

    switch (command) {	
    case 0x4141: //AA:
      if (getAACoordinates() > 0)
        return PARSER_ERROR;
      break;
    case 0x5041: //PA:
      if (getPACoordinates() > 0)
        return PARSER_ERROR;
      break;
    case 0x5044: //PD:
    case 0x5350: //SP:		// Select pen 
    case 0x5055: //PU:
      break;
    case 0x4950: //IP:		// Init point
      if (getIPCoordinates() > 0)
        return PARSER_ERROR;
      break;
    case 0x5343: //SC:		// Scale all of them processed inside parser
      if (getSCCoordinates() > 0)
        return PARSER_ERROR;
      break;
    case 0x494E: //IN:		// Home & init
      break;
    case 0x5653: //VS:
    case 0x5057: //PW:
      if (getVSValue() > 0)
        return PARSER_ERROR;
      break;
    default: 		/* Skip unknown HPGL command: */
      break;
    }

    return command;
  }


  /**
   ** getIPCoordinates() - Gets IP parameters
   **
   **/
  private int getIPCoordinates() {
    if ((numpar[IP_P1X] = getFloat())==PARSER_ERROR) {	     
      P1[X_COORD] = P1X_default;
      P1[Y_COORD] = P1Y_default;
      P2[X_COORD] = P2X_default;
      P2[Y_COORD] = P2Y_default;

      // Set S1
      S1[X_COORD] = P1[X_COORD];
      S1[Y_COORD] = P1[Y_COORD];
      S2[X_COORD] = P2[X_COORD];
      S2[Y_COORD] = P2[Y_COORD];

      Q[X_COORD] = (P2[X_COORD] - P1[X_COORD]) / (S2[X_COORD] - S1[X_COORD]);
      Q[Y_COORD] = (P2[Y_COORD] - P1[Y_COORD]) / (S2[Y_COORD] - S1[Y_COORD]);

      return 0;
    }
    if ((numpar[IP_P1Y] = getFloat())==PARSER_ERROR)	     
      return PARSER_ERROR; /* x without y! */

    if ((numpar[IP_P2X] = getFloat()) ==PARSER_ERROR) { 	/* No number found  */
      P2[X_COORD] += numpar[IP_P1X] - P1[X_COORD];
      P2[Y_COORD] += numpar[IP_P1Y] - P1[Y_COORD];

      P1[X_COORD] = numpar[IP_P1X];
      P1[Y_COORD] = numpar[IP_P1Y];
      // Set S1
      S1[X_COORD] = P1[X_COORD];
      S1[Y_COORD] = P1[Y_COORD];
      S2[X_COORD] = P2[X_COORD];
      S2[Y_COORD] = P2[Y_COORD];

      Q[X_COORD] = (P2[X_COORD] - P1[X_COORD]) / (S2[X_COORD] - S1[X_COORD]);
      Q[Y_COORD] = (P2[Y_COORD] - P1[Y_COORD]) / (S2[Y_COORD] - S1[Y_COORD]);

      return 0;
    }
    if ((numpar[IP_P2Y] = getFloat())==PARSER_ERROR)	     
      return PARSER_ERROR; /* x without y! */

    P1[X_COORD] = numpar[IP_P1X];
    P1[Y_COORD] = numpar[IP_P1Y];
    P2[X_COORD] = numpar[IP_P2X];
    P2[Y_COORD] = numpar[IP_P2Y];

    // Set S1
    S1[X_COORD] = P1[X_COORD];
    S1[Y_COORD] = P1[Y_COORD];
    S2[X_COORD] = P2[X_COORD];
    S2[Y_COORD] = P2[Y_COORD];

    Q[X_COORD] = (P2[X_COORD] - P1[X_COORD]) / (S2[X_COORD] - S1[X_COORD]);
    Q[Y_COORD] = (P2[Y_COORD] - P1[Y_COORD]) / (S2[Y_COORD] - S1[Y_COORD]);

    return 0;
  }

  /**
   ** getSCCoordinates() - Gets SC parameters
   **
   **/
  private int getSCCoordinates() {

    if ((numpar[SC_P1X] = getFloat())==PARSER_ERROR) {	     
      S1[X_COORD] = P1X_default;
      S1[Y_COORD] = P1Y_default;
      S2[X_COORD] = P2X_default;
      S2[Y_COORD] = P2Y_default;
      Q[X_COORD] = Q[Y_COORD] = 1.0;
      return 0;
    }
    if ((numpar[SC_P2X] = getFloat())==PARSER_ERROR)	     
      return PARSER_ERROR; /* x without y! */

    if ((numpar[SC_P1Y] = getFloat()) ==PARSER_ERROR) { 	/* No number found  */
      return PARSER_ERROR; /* x without y! */
    }
    if ((numpar[SC_P2Y] = getFloat())==PARSER_ERROR)	     
      return PARSER_ERROR; /* x without y! */


    if (numpar[SC_P1X] == numpar[SC_P2X] || numpar[SC_P1Y] == numpar[SC_P2Y]) { 	/* min must differ from max */
      Q[X_COORD] = Q[Y_COORD] = 1.0;
      return 0;
    }

    // Set S1

    S1[X_COORD] = numpar[SC_P1X];
    S1[Y_COORD] = numpar[SC_P1Y];

    S2[X_COORD] = numpar[SC_P2X];
    S2[Y_COORD] = numpar[SC_P2Y];

    // Update Delta

    Q[X_COORD] = (P2[X_COORD] - P1[X_COORD]) / (S2[X_COORD] - S1[X_COORD]);
    Q[Y_COORD] = (P2[Y_COORD] - P1[Y_COORD]) / (S2[Y_COORD] - S1[Y_COORD]);

    return 0;
  }


  /**
   ** getVSValue() - Gets the value of a speed command
   **/
  private int getVSValue() {
    if ((numpar[0] = getFloat())==PARSER_ERROR)	     
      return PARSER_ERROR;
    return 0;
  }
  /**
   ** getPACoordinates() - Gets the coordinates of a move command
   **/
  private int getPACoordinates()
  {

    if ((numpar[X_COORD] = getFloat())==PARSER_ERROR)	     
      return PARSER_ERROR;
    if ((numpar[Y_COORD] = getFloat())==PARSER_ERROR)	      
      return PARSER_ERROR;
    /**
     println("\ngetPACoordinates() - BEF");
     print(" numpar X: ");
     print(numpar[0]);
     print(" numpar Y ");
     println(numpar[1]);
     **/
    numpar[X_COORD] = plotter2User(numpar[X_COORD], X_COORD);
    numpar[Y_COORD] = plotter2User(numpar[Y_COORD], Y_COORD);

    return 0;
  }
  /**
   * getAACoordinates() - Gets the arc command coordinates
   **/
  private int getAACoordinates()
  {
    if ((numpar[I_COORD] = getFloat())==PARSER_ERROR)	     
      return PARSER_ERROR;
    if ((numpar[J_COORD] = getFloat())==PARSER_ERROR)	      
      return PARSER_ERROR;
    if ((numpar[ARC_THETA] = getFloat())==PARSER_ERROR)	      
      return PARSER_ERROR;
    numpar[SEG_ALPHA] = getFloat();	// Opcional


    numpar[I_COORD] = plotter2User(numpar[I_COORD], I_COORD);
    numpar[J_COORD] = plotter2User(numpar[J_COORD], J_COORD);
    /**
     // Adjust 360 arcs
     if (numpar[ARC_THETA] == 360) {
     numpar[ARC_THETA] = numpar[ARC_THETA] - .001f;
     }
     else if (numpar[ARC_THETA] == -360) {
     numpar[ARC_THETA] = numpar[ARC_THETA] + .001f;
     }
     **/
    return 0;
  }


  /**
   ** getFloat() - only accepts float positive numbers
   ** returns PARSER_ERROR if error
   **/
  private float getFloat()
  {
    char c;
    int n = 0;
    char[] numbuf =  new char[80];
    //        char numbuf[80];


    while ( (c = buffptr[buffinx++]) != '\0') {
      if (c == ' ' || c == '\t' ||c == '\n' || c== '\r')
        continue;
      if (((c >= '0') && (c <= '9')) || (c == '.') || (c == '+') || (c == '-')) {
        numbuf[n++] = c;
      } else { // end of number
        numbuf[n] = '\0';
        return (float)atof(numbuf);
      }
    }
    return PARSER_ERROR;
  }

  /**
   ** user2Plotter() - Translate user to plotter coordinates
   ** Is so basic in order to maintain compatibility with Processing.
   **/
  private float user2Plotter(float coord, int axis)
  {
    if (axis > NUM_AXIS)
      return -1;
    return (float)(P1[axis] + (coord - S1[axis]) * Q[axis])/PLOT_UNIT;
  }
  /**
   ** plotter2User() - Translate plotter to user coordinates
   ** Is so basic in order to maintain compatibility with Processing.
   **/
  private float plotter2User(float coord, int axis)
  {
    if (axis > NUM_AXIS)
      return -1;
    return  (float)(S1[axis] + (coord - P1[axis])/Q[axis])*PLOT_UNIT;
  }
  /**
   * atof() - Processing doesn´t have.
   **/
  private float atof( char[] s) { 
    String str2 = new String(s);

    return float(str2);
  }
  /**
   ** string2Char()
   **/
  private void string2Char(String s) {
    int i = 0;
    for (i = 0; i < s.length (); i++)
      mybuffer[i] = s.charAt(i);
    mybuffer[i] = '\0';
  }
}

